@@ -236,35 +236,22 @@ class MongoEventSourceRepository(
 ):
     """EventSource repository that persist event in mongodb."""
 
-    def __init__(self, client, database="fenrys", collection="event_store"):
+    def __init__(
+        self,
+        client,
+        database="fenrys",
+        collection="event_store",
+        memento_collection="domain_object_mementos",
+        logger=None,
+    ):
         super().__init__()
         self.__client = client
         self.__db = self.__client[database]
         self.__collection = self.__db[collection]
+        self.__memento_collection = self.__db[memento_collection]
+        self.__logger = logger
         self.__create_indexes()
 
-    def append_to_stream_bak(self, obj):  # noqa: D102
-        assert obj is not None
-        assert isinstance(obj, DomainObject)
-
-        max_known_version = self.max_version_for_object(obj.object_id)
-        merged_stream = self.merge_event_streams(
-            self.get_event_stream_for(obj.object_id), obj.event_stream
-        )
-        merged_stream_version = merged_stream[-1]["version"]
-
-        events_to_add = list()
-        if merged_stream_version > max_known_version:
-            for event in filter(
-                lambda e: e["version"] > max_known_version, merged_stream
-            ):
-                events_to_add.append(deepcopy(event))
-
-        if len(events_to_add) > 0:
-            self.__collection.insert_many(events_to_add)
-
-        return deepcopy(events_to_add)
-
     def append_to_stream(self, obj):  # noqa: D102
         assert obj is not None
         assert isinstance(obj, DomainObject)
@@ -273,17 +260,23 @@ class MongoEventSourceRepository(
         while True:
             try:
                 known_event_stream = self.get_event_stream_for(obj.object_id)
-                known_event_ids = []
-                max_known_version = 0
-                for event in known_event_stream:
-                    known_event_ids.append(event["event_id"])
-                    max_known_version = max(max_known_version, event["version"])
-
-                events_to_add = [
-                    event
-                    for event in obj.event_stream
-                    if event["event_id"] not in known_event_ids
+                known_event_ids = [
+                    event["event_id"] for event in known_event_stream
                 ]
+                max_known_version = (
+                    known_event_stream[-1]["version"]
+                    if len(known_event_stream) > 0
+                    else 0
+                )
+
+                events_to_add = []
+                for event in reversed(obj.event_stream):
+                    if event["event_id"] not in known_event_ids:
+                        events_to_add.append(event)
+                    else:
+                        break
+                events_to_add = list(reversed(events_to_add))
+
                 if added_events is None:
                     added_events = deepcopy(events_to_add)
 
@@ -297,12 +290,28 @@ class MongoEventSourceRepository(
             except BulkWriteError:
                 pass
 
+        memento = self.__load_memento(obj.object_id)
+        if self.__should_save_memento(memento, obj):
+            self.__save_memento(obj)
+
         return added_events
 
     def max_version_for_object(self, object_id: str) -> int:
         return 0
 
     def load(self, object_id):  # noqa: D102
+        if self.__have_memento(object_id):
+            try:
+                return self.load_with_memento(object_id)
+            except Exception as e:
+                self.__log_error(
+                    f"Cannot load memento for {object_id}: {e}. "
+                    "Removing memento and loading from stream."
+                )
+                self.__remove_memento(object_id)
+        return self.load_without_memento(object_id)
+
+    def load_without_memento(self, object_id):
         obj = self.create_blank_domain_object()
         assert isinstance(obj, DomainObject)
 
@@ -311,6 +320,16 @@ class MongoEventSourceRepository(
 
         return obj
 
+    def load_with_memento(self, object_id):
+        memento = self.__load_memento(object_id)
+        obj = self.create_blank_domain_object()
+        assert isinstance(obj, DomainObject)
+
+        stream = self.get_event_stream_for(object_id)
+        obj.rehydrate(stream, memento=memento)
+
+        return obj
+
     def exists(self, object_id):  # noqa: D102
         return self.__collection.count_documents({"object_id": object_id}) > 0
 
@@ -356,6 +375,50 @@ class MongoEventSourceRepository(
         else:
             return None
 
+    def __load_memento(self, object_id):
+        return self.__memento_collection.find_one({"object_id": object_id})
+
+    @staticmethod
+    def __should_save_memento(previous_memento, obj):
+        if obj.version_number > 100 and previous_memento is None:
+            return True
+
+        if (
+            previous_memento is not None
+            and (obj.version_number - previous_memento["version_number"]) > 100
+        ):
+            return True
+
+        return False
+
+    def __save_memento(self, obj):
+        try:
+            memento_payload = obj.get_memento()
+            if not obj.is_json_serializable(memento_payload):
+                raise ValueError(f"Memento should be json serializable")
+
+            memento = {
+                "object_id": obj.object_id,
+                "version_number": obj.version_number,
+                "memento_payload": memento_payload,
+            }
+            self.__memento_collection.replace_one(
+                {"object_id": obj.object_id}, memento, upsert=True
+            )
+        except NotImplementedError:
+            pass
+        except Exception as e:
+            self.__log_error(f"Cannot save memento : {e}")
+
+    def __remove_memento(self, object_id):
+        self.__memento_collection.delete_one({"object_id": object_id})
+
+    def __have_memento(self, object_id):
+        return (
+            self.__memento_collection.count_documents({"object_id": object_id})
+            > 0
+        )
+
     def __create_indexes(self):
         self.__collection.create_index(
             [("event_id", ASCENDING)], name="eventid_unique", unique=True
@@ -368,6 +431,15 @@ class MongoEventSourceRepository(
         self.__collection.create_index(
             [("event_timestamp", DESCENDING)], name="timestamp_ascending"
         )
+        self.__memento_collection.create_index(
+            [("object_id", ASCENDING)],
+            name="memento_objectid_unique",
+            unique=True,
+        )
+
+    def __log_error(self, message):
+        if self.__logger is not None:
+            self.__logger.error(message)
 
 
 class InMemoryEventSourceRepository(
